#|

 Gtk ffi

 Copyright (c) 2004 by Vasilis Margioulas <vasilism@sch.gr>

 You have the right to distribute and use this software as governed by 
 the terms of the Lisp Lesser GNU Public License (LLGPL):

    (http://opensource.franz.com/preamble.html)
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 Lisp Lesser GNU Public License for more details.
 
|#

(in-package :gtk-ffi)

(def-gtk-lib-functions :glib
  (g-free ((data c-pointer)))
  (g-slist-free ((lst c-pointer)))
  (g-timeout-add ((milliseconds uint)
		  (func #.(callback-function ((data c-pointer))
			    boolean))
		  (data c-pointer))
     uint)
  (g-locale-from-utf8 ((utf8-string c-pointer)
		       (len sint32)
		       (bytes-read c-pointer)
		       (bytes-written c-pointer)
		       (gerror c-pointer))
   c-string :malloc-free)
  (g-locale-to-utf8 ((local-string c-string)
		     (len sint32)
		     (bytes-read c-pointer)
		     (bytes-written c-pointer)
		     (gerror c-pointer))
   c-pointer))

(def-gtk-lib-functions :gthread
  (g-thread-init ((vtable c-pointer))))

(def-gtk-lib-functions :gdk
  (gdk-threads-init ())
  (gdk-threads-enter ())
  (gdk-threads-leave ())
  (gdk-flush ()))

(def-union g-value-data
    (v-int :int)
  (v-uint :unsigned-int)
  (v-long :long)
  (v-ulong :unsigned-long)
  (v-int64-lo :int)
  (v-int64-hi :int)
  (v-uint64-lo :unsigned-int)
  (v-uint64-hi :unsigned-int)
  (v-float :float)
  (v-double :double)
  (v-pointer :pointer-void))

(def-struct g-value
    (g-type (:array :int 16)))

(defmacro with-g-value ((var) &body body)
  `(call-with-g-value (lambda (,var) ,@body)))

(defun call-with-g-value (fn)
  (declare (optimize (speed 3) (safety 0) (space 0)))
  (let ((gva (ffx:fgn-alloc 'g-value 1 :with-g-value)))
    (unwind-protect
        (dotimes (n 16)
            (let ((gv (ffx:ff-elt gva 'g-value 0)))
              (let ((ns (get-slot-pointer gv 'g-value 'g-type)))
		(setf (deref-array ns '(:array :int) n) 0))))
          (funcall fn gva)
      (ffx:fgn-free gva))))

(eval-when (compile load eval) (export 'with-g-value))

#+test
(def-gtk-lib-functions :gobject
    (g-value-set-string ((value c-pointer)
                         (str c-string))))

#+test
(def-gtk-function :gobject g-value-set-string
  :arguments ((value c-pointer) (str c-string))
  :return-type nil :call-direct t)

(def-gtk-lib-functions :gobject
  ;; callbacks
  (g-cclosure-new ((callback-f #.(callback-function ((widget c-pointer)
						     (event c-pointer)
						     (data c-pointer))
				   boolean))
		   (user-data c-pointer)
		   (destroy-data c-pointer))
      c-pointer)
  (g-cclosure-new-swap ((callback-f #.(callback-function ((widget c-pointer)
							  (event c-pointer)
							  (data c-pointer))
				        boolean))
			(user-data c-pointer)
			(destroy-data c-pointer))
      c-pointer)
  (g-signal-connect-closure ((instance c-pointer)
                             (detailed-signal c-string)
                             (closure c-pointer)
                             (after boolean)
                             )
    ulong nil nil)
  (g-object-set-valist ((object c-pointer)
			(first-prop c-string)
			(varargs c-pointer)))
  (g-object-set-property ((object c-pointer)
                          (property-name c-string)
                          (value c-pointer)))
  (g-value-init ((value c-pointer)
                 (type ulong))
    c-pointer)
  (g-value-unset ((value c-pointer)))
  (g-value-set-string ((value c-pointer)
		       (str c-string)))
  (g-value-set-int ((value c-pointer)
		    (int int)))
  (g-value-set-long ((value c-pointer)
		     (long long)))
  (g-value-set-boolean ((value c-pointer)
			(bool boolean))
    nil nil nil)
  (g-value-set-float ((value c-pointer)
		      (float single-float)))
  (g-value-set-double ((value c-pointer)
		       (double double-float))))



